import { throwError } from "./utils";

export interface ConfigOptions {
    rsc?: {
        componentType: "client",
    }
}
export interface ParserComponentInterface {
    name: string;
    refs: ParserComponentRef[];
    state: ParserComponentState;
}

export interface ParserComponentRefInterface {
    name: string;
    used?: boolean;
    forward?: boolean;
}

export class ParserComponentRef implements ParserComponentRefInterface {
    name: string;
    used?: boolean;
    forward?: boolean;

    constructor(options: ParserComponentRefInterface) {
        this.name = options.name;
        this.used = options.used ?? false;
        this.forward = options.forward ?? false;
    }
}

export interface ParserComponentStateInterface {
    properties: ParserComponentStateProperty[];
    exists(name: string): boolean;
    get(name: string): ParserComponentStateProperty | undefined;
}

export interface ParserComponentStatePropertyInterface {
    name: string;
    derived: 'useState' | 'useStore';
    from: 'ObjectProperty' | 'ObjectMethod';
    used?: boolean;
}

class ParserComponentStateProperty implements ParserComponentStatePropertyInterface {
    name: string;
    derived: 'useState' | 'useStore';
    from: 'ObjectProperty' | 'ObjectMethod';
    used?: boolean;

    constructor(options: ParserComponentStatePropertyInterface) {
        this.name = options.name;
        this.derived = options.derived;
        this.from = options.from;
        this.used = options.used ?? false;
    }

    setUsed(used: boolean): void {
        this.used = used;
    }
}

class ParserComponentState implements ParserComponentStateInterface {
    properties: ParserComponentStateProperty[];

    constructor() {
        this.properties = [];
    }

    exists(name: string): boolean {
        return this.properties.some(property => property.name === name);
    }

    get(name: string): ParserComponentStateProperty | undefined {
        return this.properties.find(property => property.name === name);
    }
}

class ParserComponent implements ParserComponentInterface {
    name: string;
    refs: ParserComponentRef[];
    state: ParserComponentState;

    constructor(name: string) {
        this.name = name;
        this.refs = [];
        this.state = new ParserComponentState();
    }

    addRef(options: ParserComponentRefInterface): void {
        const ref = new ParserComponentRef(options);
        this.refs.push(ref);
    }

    findRef(name: string): ParserComponentRef | undefined {
        return this.refs.find(ref => ref.name === name);
    }

    hasForwardRef(): boolean {
        return this.refs.some(ref => ref.forward === true);
    }

    findForwardRef(): ParserComponentRef | undefined {
        return this.refs.find(ref => ref.forward === true);
    }

    hasRef(name: string): boolean {
        return this.refs.some(ref => ref.name === name);
    }

    addProp(options: ParserComponentStatePropertyInterface): void {
        const property = new ParserComponentStateProperty(options);
        this.state.properties.push(property);
    }

    findProp(name: string): ParserComponentStateProperty | undefined {
        return this.state.get(name);
    }

    hasProp(name: string): boolean {
        return this.state.exists(name);
    }
}

type ParserErrorMsg =
    'REF_NOT_FOUND' |
    'PROP_NOT_FOUND' |
    'PROP_SET_NOT_FOUND' |
    'COMP_NOT_FOUND' |
    'PROP_FUNC_NOT_FOUND' |
    'PROP_FUNC_COMP_NOT_FOUND';

export class ParserScope {
    components: ParserComponent[];
    cursor?: string;
    defaultForwardRefExport?: string;
    configOptions?: ConfigOptions;

    constructor() {
        this.components = [];
    }

    addComponent(name: string): void {
        const existingComponent = this.findComponent(name);
        if (!existingComponent) {
            const component = new ParserComponent(name);
            this.components.push(component);
        }
    }

    findComponent(name: string): ParserComponent | undefined {
        return this.components.find(component => component.name === name);
    }

    currentComponent(throws: boolean = false): ParserComponent | undefined {
        if (throws && !this.cursor) {
            throwError(this.errorMsg('COMP_NOT_FOUND'))
        }
        if (this.cursor) {
            return this.findComponent(this.cursor);
        }
        return undefined;
    }

    setCursor(newCursor: string | null): void {
        if (newCursor !== null && newCursor !== undefined) {
            this.cursor = newCursor;
        }
    }

    errorMsg(error: ParserErrorMsg, params?: any) {
        switch (error) {
            case 'COMP_NOT_FOUND':
                return `Current Component not found!`
            case 'REF_NOT_FOUND':
                return `ref does not exist Component: ${this.cursor} Ref: ${params.ref}!`
            case 'PROP_FUNC_NOT_FOUND':
                return `looking up state function on non-existent Component with Prop: ${params.prop}`
            case 'PROP_FUNC_COMP_NOT_FOUND':
                return `looking up state function on non-existent prop Component: ${this.cursor} Prop: ${params.prop}`
            case 'PROP_NOT_FOUND':
                return `looking up state property on non-existent Component with Prop: ${params.prop}`
            case 'PROP_SET_NOT_FOUND':
                return `current component state set without being defined: Component: ${this.cursor} prop: ${params.prop}`
        }
    }

}
